import 'package:arna/arna.dart';
import 'package:flutter/cupertino.dart' show CupertinoRouteTransitionMixin;

/// A modal route that replaces the entire screen with a transition.
///
/// By default, when a modal route is replaced by another, the previous route remains in memory. To free all the
/// resources when this is not necessary, set [maintainState] to false.
///
/// The [fullscreenDialog] property specifies whether the incoming route is a fullscreen modal dialog.
///
/// The type `T` specifies the return type of the route which can be supplied as the route is popped from the stack via
/// [Navigator.pop] by providing the optional `result` argument.
///
/// See also:
///
///  * [ArnaRouteTransitionMixin], which provides the transition for this route.
class ArnaPageRoute<T> extends PageRoute<T> with ArnaRouteTransitionMixin<T> {
  /// Construct a MaterialPageRoute whose contents are defined by [builder].
  ///
  /// The values of [builder], [maintainState], and [PageRoute.fullscreenDialog] must not be null.
  ArnaPageRoute({
    required this.builder,
    super.settings,
    this.maintainState = true,
    super.fullscreenDialog = false,
  }) {
    assert(opaque);
  }

  /// Builds the primary contents of the route.
  final WidgetBuilder builder;

  @override
  Widget buildContent(BuildContext context) => builder(context);

  @override
  final bool maintainState;

  @override
  String get debugLabel => '${super.debugLabel}(${settings.name})';
}

/// A mixin that provides transitions for a [PageRoute].
mixin ArnaRouteTransitionMixin<T> on PageRoute<T> {
  /// Builds the primary contents of the route.
  @protected
  Widget buildContent(BuildContext context);

  @override
  Duration get transitionDuration => Styles.routeDuration;

  @override
  Color? get barrierColor => null;

  @override
  String? get barrierLabel => null;

  @override
  bool canTransitionTo(TransitionRoute<dynamic> nextRoute) {
    // Don't perform outgoing animation if the next route is a fullscreen dialog.
    return nextRoute is ArnaRouteTransitionMixin && !nextRoute.fullscreenDialog;
  }

  @override
  Widget buildPage(BuildContext context, Animation<double> animation, Animation<double> secondaryAnimation) {
    final Widget result = buildContent(context);
    assert(() {
      // ignore: unnecessary_null_comparison
      if (result == null) {
        throw FlutterError(
          'The builder for route "${settings.name}" returned null.\n'
          'Route builders must never return null.',
        );
      }
      return true;
    }());
    return Semantics(
      scopesRoute: true,
      explicitChildNodes: true,
      child: result,
    );
  }

  @override
  Widget buildTransitions(
    BuildContext context,
    Animation<double> animation,
    Animation<double> secondaryAnimation,
    Widget child,
  ) {
    return CupertinoRouteTransitionMixin.buildPageTransitions<T>(this, context, animation, secondaryAnimation, child);
  }
}
